#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#















"""Tests for key.py."""

import base64
import collections
import datetime
import os
import pickle
import sortedcontainers

from google.appengine.ext.ndb import key
from google.appengine.ext.ndb import model
from google.appengine.ext.ndb import tasklets
from google.appengine.ext.ndb import test_utils
import six
from six.moves import range
from six.moves import zip

from google.appengine.api import datastore_errors
from google.appengine.api import datastore_types
from google.appengine.datastore import entity_bytes_pb2 as entity_pb2
from absl.testing import absltest as unittest


class KeyTests(test_utils.NDBTest):

  the_module = key

  def testShort(self):
    k0 = key.Key('Kind', None)
    self.assertEqual(k0.flat(), ('Kind', None))
    k1 = key.Key('Kind', 1)
    self.assertEqual(k1.flat(), ('Kind', 1))
    k2 = key.Key('Parent', 42, 'Kind', 1)
    self.assertEqual(k2.flat(), ('Parent', 42, 'Kind', 1))

  def testFlat(self):
    flat = ('Kind', 1)
    pairs = tuple((flat[i], flat[i + 1]) for i in range(0, len(flat), 2))
    k = key.Key(flat=flat)
    self.assertEqual(k.pairs(), pairs)
    self.assertEqual(k.flat(), flat)
    self.assertEqual(k.kind(), 'Kind')

  def testFlatLong(self):
    flat = ('Kind', 1, 'Subkind', 'foobar')
    pairs = tuple((flat[i], flat[i + 1]) for i in range(0, len(flat), 2))
    k = key.Key(flat=flat)
    self.assertEqual(k.pairs(), pairs)
    self.assertEqual(k.flat(), flat)
    self.assertEqual(k.kind(), 'Subkind')

  def testSerialized(self):
    flat = ['Kind', 1, 'Subkind', 'foobar']
    r = entity_pb2.Reference()
    r.app = 'ndb-test-app-id'
    e = r.path.element.add()
    e.type = flat[0]
    e.id = flat[1]
    e = r.path.element.add()
    e.type = flat[2]
    e.name = flat[3]
    serialized = r.SerializeToString()
    ref_bytes = six.ensure_binary(r.SerializeToString())
    urlsafe = base64.urlsafe_b64encode(ref_bytes).rstrip(b'=')

    k = key.Key(flat=flat)
    self.assertEqual(k.serialized(), serialized)
    self.assertEqual(k.urlsafe(), urlsafe)
    self.assertEqual(k.reference(), r)

    k = key.Key(urlsafe=urlsafe)
    self.assertEqual(k.serialized(), serialized)
    self.assertEqual(k.urlsafe(), urlsafe)
    self.assertEqual(k.reference(), r)

    k = key.Key(serialized=serialized)
    self.assertEqual(k.serialized(), serialized)
    self.assertEqual(k.urlsafe(), urlsafe)
    self.assertEqual(k.reference(), r)

    k = key.Key(reference=r)
    self.assertIsNot(k.reference(), r)
    self.assertEqual(k.serialized(), serialized)
    self.assertEqual(k.urlsafe(), urlsafe)
    self.assertEqual(k.reference(), r)

    k = key.Key(reference=r, app=r.app, namespace='')
    self.assertIsNot(k.reference(), r)
    self.assertEqual(k.serialized(), serialized)
    self.assertEqual(k.urlsafe(), urlsafe)
    self.assertEqual(k.reference(), r)

    k1 = key.Key('A', 1)
    self.assertEqual(k1.urlsafe(), b'ag9uZGItdGVzdC1hcHAtaWRyBwsSAUEYAQw')
    k2 = key.Key(urlsafe=k1.urlsafe())
    self.assertEqual(k1, k2)

  def testId(self):
    k1 = key.Key('Kind', 'foo', app='app1', namespace='ns1')
    self.assertEqual(k1.id(), 'foo')

    k2 = key.Key('Subkind', 42, parent=k1)
    self.assertEqual(k2.id(), 42)

    k3 = key.Key('Subkind', 'bar', parent=k2)
    self.assertEqual(k3.id(), 'bar')


    k4 = key.Key('Subkind', None, parent=k3)
    self.assertEqual(k4.id(), None)

  def testIdentity(self):
    test_kind, test_id = 'test-kind', 'test-id'
    k = key.Key(test_kind, test_id)

    with self.subTest(name='Kind'):
      self.assertEqual(k.kind(), test_kind)

    with self.subTest(name='ID'):
      self.assertEqual(k.id(), test_id)

  def testStringId(self):
    k1 = key.Key('Kind', 'foo', app='app1', namespace='ns1')
    self.assertEqual(k1.string_id(), 'foo')

    k2 = key.Key('Subkind', 'bar', parent=k1)
    self.assertEqual(k2.string_id(), 'bar')

    k3 = key.Key('Subkind', 42, parent=k2)
    self.assertEqual(k3.string_id(), None)


    k4 = key.Key('Subkind', None, parent=k3)
    self.assertEqual(k4.string_id(), None)

  def testIntegerId(self):
    k1 = key.Key('Kind', 42, app='app1', namespace='ns1')
    self.assertEqual(k1.integer_id(), 42)

    k2 = key.Key('Subkind', 43, parent=k1)
    self.assertEqual(k2.integer_id(), 43)

    k3 = key.Key('Subkind', 'foobar', parent=k2)
    self.assertEqual(k3.integer_id(), None)


    k4 = key.Key('Subkind', None, parent=k3)
    self.assertEqual(k4.integer_id(), None)

  def testParent(self):
    p = key.Key('Kind', 1, app='app1', namespace='ns1')
    self.assertEqual(p.parent(), None)

    k = key.Key('Subkind', 'foobar', parent=p)
    self.assertEqual(k.flat(), ('Kind', 1, 'Subkind', 'foobar'))
    self.assertEqual(k.parent(), p)

    k = key.Key(
        'Subkind', 'foobar', parent=p, app=p.app(), namespace=p.namespace())
    self.assertEqual(k.flat(), ('Kind', 1, 'Subkind', 'foobar'))
    self.assertEqual(k.parent(), p)

  def testRoot(self):
    p = key.Key('Kind', 1, app='app1', namespace='ns1')
    self.assertEqual(p.root(), p)

    k = key.Key('Subkind', 'foobar', parent=p)
    self.assertEqual(k.flat(), ('Kind', 1, 'Subkind', 'foobar'))
    self.assertEqual(k.root(), p)

    k2 = key.Key(
        'Subsubkind', 42, parent=k, app=p.app(), namespace=p.namespace())
    self.assertEqual(k2.flat(),
                     ('Kind', 1, 'Subkind', 'foobar', 'Subsubkind', 42))
    self.assertEqual(k2.root(), p)

  def testRepr_Inferior(self):
    k = key.Key('Kind', 1, 'Subkind', 'foobar')
    self.assertEqual(repr(k), "Key('Kind', 1, 'Subkind', 'foobar')")
    self.assertEqual(repr(k), str(k))

  def testRepr_Toplevel(self):
    k = key.Key('Kind', 1)
    self.assertEqual(repr(k), "Key('Kind', 1)")

  def testRepr_Incomplete(self):
    k = key.Key('Kind', None)
    self.assertEqual(repr(k), "Key('Kind', None)")

  def testRepr_UnicodeKind(self):
    k = key.Key(u'\u1234', 1)
    if six.PY2:
      self.assertEqual(repr(k), "Key('\\xe1\\x88\\xb4', 1)")
    else:
      self.assertEqual(repr(k), u"Key('\u1234', 1)")

  def testRepr_UnicodeId(self):
    k = key.Key('Kind', u'\u1234')
    if six.PY2:
      self.assertEqual(repr(k), "Key('Kind', '\\xe1\\x88\\xb4')")
    else:
      self.assertEqual(repr(k), u"Key('Kind', '\u1234')")

  def testRepr_App(self):
    k = key.Key('Kind', 1, app='foo')
    self.assertEqual(repr(k), "Key('Kind', 1, app='foo')")

  def testRepr_Namespace(self):
    k = key.Key('Kind', 1, namespace='foo')
    self.assertEqual(repr(k), "Key('Kind', 1, namespace='foo')")

  def testUnicode(self):
    flat_input = (u'Kind\u1234', 1, 'Subkind', u'foobar\u4321')
    flat = (six.ensure_str(flat_input[0]), flat_input[1], flat_input[2],
            six.ensure_str(flat_input[3]))
    pairs = tuple((flat[i], flat[i + 1]) for i in range(0, len(flat), 2))
    k = key.Key(flat=flat_input)
    self.assertEqual(k.pairs(), pairs)
    self.assertEqual(k.flat(), flat)

    r = k.reference()
    serialized = k.serialized()
    urlsafe = k.urlsafe()
    key.Key(urlsafe=urlsafe.decode('utf8'))
    key.Key(serialized=serialized.decode('utf8'))
    key.Key(reference=r)

    r = entity_pb2.Reference()
    r.app = 'ndb-test-app-id'
    e = r.path.element.add()
    e.type = flat[0]
    e.name = flat[3]
    k = key.Key(reference=r)
    self.assertEqual(k.reference(), r)

  def testHash(self):
    flat = ['Kind', 1, 'Subkind', 'foobar']
    pairs = [(flat[i], flat[i + 1]) for i in range(0, len(flat), 2)]
    k = key.Key(flat=flat)
    self.assertEqual(hash(k), hash(tuple(pairs)))

  def testOrdering(self):

    a = key.Key(app='app2', namespace='ns2', flat=('kind1', 1))
    b = key.Key(app='app2', namespace='ns1', flat=('kind1', 1))
    c = key.Key(app='app1', namespace='ns1', flat=('kind1', 1))
    d = key.Key(app='app1', namespace='ns1', flat=('kind1', 2))
    e = key.Key(app='app1', namespace='ns1', flat=('kind1', 'e'))
    f = key.Key(app='app1', namespace='ns1', flat=('kind1', 'f'))
    g = key.Key(app='app1', namespace='ns1', flat=('kind2', 'f', 'x', 1))
    h = key.Key(app='app1', namespace='ns1', flat=('kind2', 'f', 'x', 2))

    expected = [c, d, e, f, g, h, b, a]
    actual = sorted([a, b, c, d, e, f, g, h])
    self.assertEqual(actual, expected)

    for i in range(len(actual)):
      for j in range(len(actual)):
        self.assertEqual(actual[i] < actual[j], i < j)
        self.assertEqual(actual[i] <= actual[j], i <= j)
        self.assertEqual(actual[i] > actual[j], i > j)
        self.assertEqual(actual[i] >= actual[j], i >= j)
        self.assertEqual(actual[i] == actual[j], i == j)
        self.assertEqual(actual[i] != actual[j], i != j)

  def testUniqueIncomplete(self):
    p0 = None
    p1 = key.Key('bar', 1)
    for p in p0, p1:
      a = key.Key('foo', 0, parent=p)
      b = key.Key('foo', '', parent=p)
      c = key.Key('foo', None, parent=p)
      self.assertEqual(a, b)
      self.assertEqual(b, c)
      self.assertEqual(c, a)
      for x in a, b, c:
        self.assertEqual(x.id(), None)
        self.assertEqual(x.string_id(), None)
        self.assertEqual(x.integer_id(), None)
        self.assertEqual(x.pairs()[-1], ('foo', None))
        self.assertEqual(x.flat()[-1], None)
        self.assertEqual(x.urlsafe(), c.urlsafe())

  def testIncomplete(self):
    key.Key(flat=['Kind', None])
    self.assertRaises(
        datastore_errors.BadArgumentError,
        key.Key,
        flat=['Kind', None, 'Subkind', 1])
    self.assertRaises(TypeError, key.Key, flat=['Kind', ()])

  def testKindFromModel(self):

    class M(model.Model):
      pass

    class N(model.Model):

      @classmethod
      def _get_kind(cls):
        return 'NN'

    k = key.Key(M, 1)
    self.assertEqual(k, key.Key('M', 1))
    k = key.Key('X', 1, N, 2, 'Y', 3)
    self.assertEqual(k, key.Key('X', 1, 'NN', 2, 'Y', 3))

  def testKindFromBadValue(self):

    self.assertRaises(Exception, key.Key, 42, 42)

  def testDeleteHooksCalled(self):
    test = self
    self.pre_counter = 0
    self.post_counter = 0

    class HatStand(model.Model):

      @classmethod
      def _pre_delete_hook(cls, key):
        test.pre_counter += 1
        if test.pre_counter == 1:
          self.assertEqual(self.key, key)

      @classmethod
      def _post_delete_hook(cls, key, future):
        test.post_counter += 1
        self.assertEqual(self.key, key)
        self.assertIs(future.get_result(), None)

    furniture = HatStand()
    key = furniture.put()
    self.key = key
    self.assertEqual(self.pre_counter, 0, 'Pre delete hook called early')
    future = key.delete_async()
    self.assertEqual(self.pre_counter, 1, 'Pre delete hook not called')
    self.assertEqual(self.post_counter, 0, 'Post delete hook called early')
    future.get_result()
    self.assertEqual(self.post_counter, 1, 'Post delete hook not called')


    new_furniture = [HatStand() for _ in range(10)]
    keys = [furniture.put() for furniture in new_furniture]
    multi_future = model.delete_multi_async(keys)
    self.assertEqual(self.pre_counter, 11,
                     'Pre delete hooks not called on delete_multi')
    self.assertEqual(self.post_counter, 1,
                     'Post delete hooks called early on delete_multi')
    for fut, key in zip(multi_future, keys):
      self.key = key
      fut.get_result()
    self.assertEqual(self.post_counter, 11,
                     'Post delete hooks not called on delete_multi')

  def testNoDefaultDeleteCallback(self):

    ctx = tasklets.get_context()
    ctx.set_cache_policy(False)

    class EmptyModel(model.Model):
      pass

    entity = EmptyModel()
    entity.put()
    fut = entity.key.delete_async()
    self.assertFalse(fut._immediate_callbacks,
                     'Delete hook queued default no-op.')

  def testGetHooksCalled(self):
    test = self
    self.pre_counter = 0
    self.post_counter = 0

    class HatStand(model.Model):

      @classmethod
      def _pre_get_hook(cls, key):
        test.pre_counter += 1
        if test.pre_counter == 1:
          self.assertEqual(key, self.key)

      @classmethod
      def _post_get_hook(cls, key, future):
        test.post_counter += 1
        self.assertEqual(key, self.key)
        self.assertEqual(future.get_result(), self.entity)

    furniture = HatStand()
    self.entity = furniture
    key = furniture.put()
    self.key = key
    self.assertEqual(self.pre_counter, 0, 'Pre get hook called early')
    future = key.get_async()
    self.assertEqual(self.pre_counter, 1, 'Pre get hook not called')
    self.assertEqual(self.post_counter, 0, 'Post get hook called early')
    future.get_result()
    self.assertEqual(self.post_counter, 1, 'Post get hook not called')


    new_furniture = [HatStand() for _ in range(10)]
    keys = [furniture.put() for furniture in new_furniture]
    multi_future = model.get_multi_async(keys)
    self.assertEqual(self.pre_counter, 11,
                     'Pre get hooks not called on get_multi')
    self.assertEqual(self.post_counter, 1,
                     'Post get hooks called early on get_multi')
    for fut, key, entity in zip(multi_future, keys, new_furniture):
      self.key = key
      self.entity = entity
      fut.get_result()
    self.assertEqual(self.post_counter, 11,
                     'Post get hooks not called on get_multi')

  def testMonkeyPatchHooks(self):
    hook_attr_names = ('_pre_get_hook', '_post_get_hook', '_pre_delete_hook',
                       '_post_delete_hook')
    original_hooks = {}


    for name in hook_attr_names:
      original_hooks[name] = getattr(model.Model, name)

    self.pre_get_flag = False
    self.post_get_flag = False
    self.pre_delete_flag = False
    self.post_delete_flag = False


    class HatStand(model.Model):

      @classmethod
      def _pre_get_hook(cls, unused_key):
        self.pre_get_flag = True

      @classmethod
      def _post_get_hook(cls, unused_key, unused_future):
        self.post_get_flag = True

      @classmethod
      def _pre_delete_hook(cls, unused_key):
        self.pre_delete_flag = True

      @classmethod
      def _post_delete_hook(cls, unused_key, unused_future):
        self.post_delete_flag = True


    for name in hook_attr_names:
      hook = getattr(HatStand, name)
      setattr(model.Model, name, hook)

    try:
      key = HatStand().put()
      key.get()
      self.assertTrue(self.pre_get_flag,
                      'Pre get hook not called when model is monkey patched')
      self.assertTrue(self.post_get_flag,
                      'Post get hook not called when model is monkey patched')
      key.delete()
      self.assertTrue(
          self.pre_delete_flag,
          'Pre delete hook not called when model is monkey patched')
      self.assertTrue(
          self.post_delete_flag, 'Post delete hook not called when model '
          'is monkey patched')
    finally:

      for name in hook_attr_names:
        setattr(model.Model, name, original_hooks[name])

  def testPreHooksCannotCancelRPC(self):

    class Foo(model.Model):

      @classmethod
      def _pre_get_hook(cls, unused_key):
        raise tasklets.Return()

      @classmethod
      def _pre_delete_hook(cls, unused_key):
        raise tasklets.Return()

    entity = Foo()
    entity.put()
    self.assertRaises(tasklets.Return, entity.key.get)
    self.assertRaises(tasklets.Return, entity.key.delete)

  def testNoDefaultGetCallback(self):

    ctx = tasklets.get_context()
    ctx.set_cache_policy(False)

    class EmptyModel(model.Model):
      pass

    entity = EmptyModel()
    entity.put()
    fut = entity.key.get_async()
    self.assertFalse(fut._immediate_callbacks, 'Get hook queued default no-op.')

  def testFromOldKey(self):
    old_key = datastore_types.Key.from_path('TestKey', 1234)
    new_key = key.Key.from_old_key(old_key)
    self.assertEqual(str(old_key), six.ensure_text(new_key.urlsafe()))

    old_key2 = new_key.to_old_key()
    self.assertEqual(old_key, old_key2)



Snapshot = collections.namedtuple('Snapshot', ['snapshot_key', 'created_on'])


class KeyPickleTests(test_utils.NDBTest):
  """Tests for key pickling."""

  def setUp(self):
    super(KeyPickleTests, self).setUp()
    self.keys = [
        key.Key(flat=['Kind', 1]),
        key.Key(flat=['Kind', 1, 'Subkind', 'foobar']),
        key.Key(
            flat=['Kind', 1, 'Subkind', 'foobar'],
            namespace='ns',
            app='a-different-app')
    ]



    self.pkeys = [
        [
            b"ccopy_reg\n_reconstructor\np0\n(cgoogle.appengine.ext.ndb.key\nKey\np1\nc__builtin__\nobject\np2\nNtp3\nRp4\n((dp5\nS'namespace'\np6\nS''\np7\nsS'app'\np8\nS'ndb-test-app-id'\np9\nsS'pairs'\np10\n(lp11\n(S'Kind'\np12\nI1\ntp13\nastp14\nb.",
            b"ccopy_reg\n_reconstructor\np0\n(cgoogle.appengine.ext.ndb.key\nKey\np1\nc__builtin__\nobject\np2\nNtp3\nRp4\n((dp5\nS'namespace'\np6\nS''\np7\nsS'app'\np8\nS'ndb-test-app-id'\np9\nsS'pairs'\np10\n(lp11\n(S'Kind'\np12\nI1\ntp13\na(S'Subkind'\np14\nS'foobar'\np15\ntp16\nastp17\nb.",
            b"ccopy_reg\n_reconstructor\np0\n(cgoogle.appengine.ext.ndb.key\nKey\np1\nc__builtin__\nobject\np2\nNtp3\nRp4\n((dp5\nS'namespace'\np6\nS'ns'\np7\nsS'app'\np8\nS'a-different-app'\np9\nsS'pairs'\np9\n(lp10\n(S'Kind'\np11\nI1\ntp12\na(S'Subkind'\np13\nS'foobar'\np14\ntp15\nastp16\nb."
        ],
        [
            b"ccopy_reg\n_reconstructor\np0\n(cgoogle.appengine.ext.ndb.key\nKey\np1\nc__builtin__\nobject\np2\nNtp3\nRp4\n((dp5\nS'app'\np6\nS'ndb-test-app-id'\np7\nsS'pairs'\np8\n((S'Kind'\np9\nI1\ntp10\ntp11\nsS'namespace'\np12\nS''\np13\nstp14\nb.",
            b"ccopy_reg\n_reconstructor\np0\n(cgoogle.appengine.ext.ndb.key\nKey\np1\nc__builtin__\nobject\np2\nNtp3\nRp4\n((dp5\nS'app'\np6\nS'ndb-test-app-id'\np7\nsS'pairs'\np8\n((S'Kind'\np9\nI1\ntp10\n(S'Subkind'\np11\nS'foobar'\np12\ntp13\ntp14\nsS'namespace'\np15\nS''\np16\nstp17\nb.",
            b"ccopy_reg\n_reconstructor\np0\n(cgoogle.appengine.ext.ndb.key\nKey\np1\nc__builtin__\nobject\np2\nNtp3\nRp4\n((dp5\nS'app'\np6\nS'a-different-app'\np7\nsS'pairs'\np8\n((S'Kind'\np9\nI1\ntp10\n(S'Subkind'\np11\nS'foobar'\np12\ntp13\ntp14\nsS'namespace'\np15\nS'ns'\np16\nstp17\nb."
        ]
    ]

  def _Unpickle(self, s):
    if six.PY2:
      return pickle.loads(s)
    else:
      return pickle.loads(s, encoding='bytes')

  def testPickleBackwardsCompatibility(self):
    for pkey_list in self.pkeys:
      for expected, pkey in zip(self.keys, pkey_list):
        expected.app()
        actual = self._Unpickle(pkey)
        self.assertEqual(expected, actual)

  def testPickleForwardCompatibility(self):
    os.environ['NDB_PY2_UNPICKLE_COMPAT'] = '1'
    self.addCleanup(os.unsetenv, 'NDB_PY2_UNPICKLE_COMPAT')
    expected = sortedcontainers.SortedSet([
        Snapshot(
            snapshot_key=key.Key('Kind', 1, 'Subkind', 'foobar', app='test'),
            created_on=datetime.datetime(2020, 8, 25, 18, 28, 27, 66651))
    ])
    py2_pickle = b'\x80\x02csortedcontainers.sortedset\nSortedSet\nq\x00c__builtin__\nset\nq\x01]q\x02cgoogle.appengine.ext.ndb.key_test\nSnapshot\nq\x03cgoogle.appengine.ext.ndb.key\nKey\nq\x04}q\x05(U\x05pairsq\x06U\x04Kindq\x07K\x01\x86q\x08U\x07Subkindq\tU\x06foobarq\n\x86q\x0b\x86q\x0cU\tnamespaceq\rU\x00q\x0eU\x03appq\x0fU\x04testq\x10u\x85q\x11\x81q\x12}q\x13(h\x06h\x0ch\rh\x0eh\x0fh\x10u\x85q\x14bcdatetime\ndatetime\nq\x15U\n\x07\xe4\x08\x19\x12\x1c\x1b\x01\x04[q\x16\x85q\x17Rq\x18\x86q\x19\x81q\x1aa\x85q\x1bRq\x1cN\x86q\x1dRq\x1e.'
    actual = self._Unpickle(six.ensure_binary(py2_pickle))
    self.assertEqual(expected, actual)

  def testPy2PickleSetState(self):

    pickle_bytes = b'\x80\x02cgoogle.appengine.ext.ndb.key\nKey\nq\x00}q\x01(U\x09namespaceq\x02U\x00q\x03U\x03appq\x04U\x0ds~some-app-idq\x05U\x05pairsq\x06U\x05Classq\x07U\x02idq\x08\x86q\x09\x85q\nu\x85q\x0b\x81q\x0c}q\x0d(h\x02h\x03h\x04h\x05h\x06h\nu\x85q\x0eb.'

    expected = key.Key('Class', 'id', app='s~some-app-id')

    actual = pickle.loads(pickle_bytes)
    self.assertEqual(expected.__getstate__()[0]['pairs'],
                     actual.__getstate__()[0]['pairs'])




  def testConsistentPickleBehaviour(self):
    k = key.Key('Kind', 'foo', app='app1', namespace='ns1')
    k.__setstate__(
        ({'pairs': (('Kind', 'foo'),), 'app': 'app1', 'namespace': 'ns1'},)
    )
    for protocol in range(pickle.HIGHEST_PROTOCOL + 1):
      self.assertEqual(k, self._Unpickle(pickle.dumps(k, protocol)))

  def testPickling(self):
    for k in self.keys:
      for protocol in range(pickle.HIGHEST_PROTOCOL + 1):
        self.assertEqual(k, self._Unpickle(pickle.dumps(k, protocol)))


def _bytes2str(bytestring):
  """Helper method for debugging pickle.dumps output."""
  return ''.join(_human_readable_byte(x) for x in bytestring)


def _human_readable_byte(c):
  if isinstance(c, str):
    c = ord(c)
  if c == 10:
    return '\\n'
  if c > 21 and c < 127:
    return chr(c)
  else:
    return '\\x{:02x}'.format(c)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#



"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n-google/appengine/datastore/entity_bytes.proto\x12\x19storage_onestore_v3_bytes\"\xe9\x05\n\rPropertyValue\x12\x12\n\nint64Value\x18\x01 \x01(\x03\x12\x14\n\x0c\x62ooleanValue\x18\x02 \x01(\x08\x12\x13\n\x0bstringValue\x18\x03 \x01(\x0c\x12\x13\n\x0b\x64oubleValue\x18\x04 \x01(\x01\x12G\n\npointvalue\x18\x05 \x01(\n23.storage_onestore_v3_bytes.PropertyValue.PointValue\x12\x45\n\tuservalue\x18\x08 \x01(\n22.storage_onestore_v3_bytes.PropertyValue.UserValue\x12O\n\x0ereferencevalue\x18\x0c \x01(\n27.storage_onestore_v3_bytes.PropertyValue.ReferenceValue\x1a\"\n\nPointValue\x12\t\n\x01x\x18\x06 \x02(\x01\x12\t\n\x01y\x18\x07 \x02(\x01\x1a\xa4\x01\n\tUserValue\x12\r\n\x05\x65mail\x18\t \x02(\t\x12\x13\n\x0b\x61uth_domain\x18\n \x02(\t\x12\x10\n\x08nickname\x18\x0b \x01(\t\x12\x0e\n\x06gaiaid\x18\x12 \x02(\x03\x12\x19\n\x11obfuscated_gaiaid\x18\x13 \x01(\t\x12\x1a\n\x12\x66\x65\x64\x65rated_identity\x18\x15 \x01(\t\x12\x1a\n\x12\x66\x65\x64\x65rated_provider\x18\x16 \x01(\t\x1a\xd7\x01\n\x0eReferenceValue\x12\x0b\n\x03\x61pp\x18\r \x02(\t\x12\x12\n\nname_space\x18\x14 \x01(\t\x12X\n\x0bpathelement\x18\x0e \x03(\n2C.storage_onestore_v3_bytes.PropertyValue.ReferenceValue.PathElement\x12\x13\n\x0b\x64\x61tabase_id\x18\x17 \x01(\t\x1a\x35\n\x0bPathElement\x12\x0c\n\x04type\x18\x0f \x02(\t\x12\n\n\x02id\x18\x10 \x01(\x03\x12\x0c\n\x04name\x18\x11 \x01(\t\"\xc8\x04\n\x08Property\x12H\n\x07meaning\x18\x01 \x01(\x0e\x32+.storage_onestore_v3_bytes.Property.Meaning:\nNO_MEANING\x12\x13\n\x0bmeaning_uri\x18\x02 \x01(\t\x12\x0c\n\x04name\x18\x03 \x02(\t\x12\x37\n\x05value\x18\x05 \x02(\x0b\x32(.storage_onestore_v3_bytes.PropertyValue\x12\x10\n\x08multiple\x18\x04 \x02(\x08\x12\x13\n\x07stashed\x18\x06 \x01(\x05:\x02-1\x12\x17\n\x08\x63omputed\x18\x07 \x01(\x08:\x05\x66\x61lse\"\xd5\x02\n\x07Meaning\x12\x0e\n\nNO_MEANING\x10\x00\x12\x08\n\x04\x42LOB\x10\x0e\x12\x08\n\x04TEXT\x10\x0f\x12\x0e\n\nBYTESTRING\x10\x10\x12\x11\n\rATOM_CATEGORY\x10\x01\x12\r\n\tATOM_LINK\x10\x02\x12\x0e\n\nATOM_TITLE\x10\x03\x12\x10\n\x0c\x41TOM_CONTENT\x10\x04\x12\x10\n\x0c\x41TOM_SUMMARY\x10\x05\x12\x0f\n\x0b\x41TOM_AUTHOR\x10\x06\x12\x0b\n\x07GD_WHEN\x10\x07\x12\x0c\n\x08GD_EMAIL\x10\x08\x12\x10\n\x0cGEORSS_POINT\x10\t\x12\t\n\x05GD_IM\x10\n\x12\x12\n\x0eGD_PHONENUMBER\x10\x0b\x12\x14\n\x10GD_POSTALADDRESS\x10\x0c\x12\r\n\tGD_RATING\x10\r\x12\x0b\n\x07\x42LOBKEY\x10\x11\x12\x10\n\x0c\x45NTITY_PROTO\x10\x13\x12\x0e\n\nEMPTY_LIST\x10\x18\x12\x0f\n\x0bINDEX_VALUE\x10\x12\"s\n\x04Path\x12\x38\n\x07\x65lement\x18\x01 \x03(\n2\'.storage_onestore_v3_bytes.Path.Element\x1a\x31\n\x07\x45lement\x12\x0c\n\x04type\x18\x02 \x02(\t\x12\n\n\x02id\x18\x03 \x01(\x03\x12\x0c\n\x04name\x18\x04 \x01(\t\"p\n\tReference\x12\x0b\n\x03\x61pp\x18\r \x02(\t\x12\x12\n\nname_space\x18\x14 \x01(\t\x12-\n\x04path\x18\x0e \x02(\x0b\x32\x1f.storage_onestore_v3_bytes.Path\x12\x13\n\x0b\x64\x61tabase_id\x18\x17 \x01(\t\"\x9f\x01\n\x04User\x12\r\n\x05\x65mail\x18\x01 \x02(\t\x12\x13\n\x0b\x61uth_domain\x18\x02 \x02(\t\x12\x10\n\x08nickname\x18\x03 \x01(\t\x12\x0e\n\x06gaiaid\x18\x04 \x02(\x03\x12\x19\n\x11obfuscated_gaiaid\x18\x05 \x01(\t\x12\x1a\n\x12\x66\x65\x64\x65rated_identity\x18\x06 \x01(\t\x12\x1a\n\x12\x66\x65\x64\x65rated_provider\x18\x07 \x01(\t\"\x9c\x03\n\x0b\x45ntityProto\x12\x31\n\x03key\x18\r \x02(\x0b\x32$.storage_onestore_v3_bytes.Reference\x12\x35\n\x0c\x65ntity_group\x18\x10 \x02(\x0b\x32\x1f.storage_onestore_v3_bytes.Path\x12.\n\x05owner\x18\x11 \x01(\x0b\x32\x1f.storage_onestore_v3_bytes.User\x12\x39\n\x04kind\x18\x04 \x01(\x0e\x32+.storage_onestore_v3_bytes.EntityProto.Kind\x12\x10\n\x08kind_uri\x18\x05 \x01(\t\x12\x35\n\x08property\x18\x0e \x03(\x0b\x32#.storage_onestore_v3_bytes.Property\x12\x39\n\x0craw_property\x18\x0f \x03(\x0b\x32#.storage_onestore_v3_bytes.Property\"4\n\x04Kind\x12\x0e\n\nGD_CONTACT\x10\x01\x12\x0c\n\x08GD_EVENT\x10\x02\x12\x0e\n\nGD_MESSAGE\x10\x03\"B\n\x0e\x45ntityMetadata\x12\x17\n\x0f\x63reated_version\x18\x01 \x01(\x03\x12\x17\n\x0fupdated_version\x18\x02 \x01(\x03\"\xbb\x01\n\rEntitySummary\x12T\n\x12large_raw_property\x18\x01 \x03(\x0b\x32\x38.storage_onestore_v3_bytes.EntitySummary.PropertySummary\x1aT\n\x0fPropertySummary\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\x1f\n\x17property_type_for_stats\x18\x02 \x01(\t\x12\x12\n\nsize_bytes\x18\x03 \x01(\x05\"4\n\x11\x43ompositeProperty\x12\x10\n\x08index_id\x18\x01 \x02(\x03\x12\r\n\x05value\x18\x02 \x03(\x0c\"\xda\x04\n\x05Index\x12\x13\n\x0b\x65ntity_type\x18\x01 \x02(\t\x12\x10\n\x08\x61ncestor\x18\x05 \x02(\x08\x12\x0e\n\x06parent\x18\x07 \x01(\x08\x12N\n\x07version\x18\x08 \x01(\x0e\x32(.storage_onestore_v3_bytes.Index.Version:\x13VERSION_UNSPECIFIED\x12;\n\x08property\x18\x02 \x03(\n2).storage_onestore_v3_bytes.Index.Property\x1a\xd0\x02\n\x08Property\x12\x0c\n\x04name\x18\x03 \x02(\t\x12]\n\tdirection\x18\x04 \x01(\x0e\x32\x33.storage_onestore_v3_bytes.Index.Property.Direction:\x15\x44IRECTION_UNSPECIFIED\x12N\n\x04mode\x18\x06 \x01(\x0e\x32..storage_onestore_v3_bytes.Index.Property.Mode:\x10MODE_UNSPECIFIED\"E\n\tDirection\x12\x19\n\x15\x44IRECTION_UNSPECIFIED\x10\x00\x12\r\n\tASCENDING\x10\x01\x12\x0e\n\nDESCENDING\x10\x02\"@\n\x04Mode\x12\x14\n\x10MODE_UNSPECIFIED\x10\x00\x12\x0e\n\nGEOSPATIAL\x10\x03\x12\x12\n\x0e\x41RRAY_CONTAINS\x10\x04\":\n\x07Version\x12\x17\n\x13VERSION_UNSPECIFIED\x10\x00\x12\x06\n\x02V1\x10\x01\x12\x06\n\x02V2\x10\x02\x12\x06\n\x02V3\x10\x03\"\xc0\x04\n\x0e\x43ompositeIndex\x12\x0e\n\x06\x61pp_id\x18\x01 \x02(\t\x12\x13\n\x0b\x64\x61tabase_id\x18\x0c \x01(\t\x12\n\n\x02id\x18\x02 \x02(\x03\x12\x34\n\ndefinition\x18\x03 \x02(\x0b\x32 .storage_onestore_v3_bytes.Index\x12>\n\x05state\x18\x04 \x02(\x0e\x32/.storage_onestore_v3_bytes.CompositeIndex.State\x12S\n\x0eworkflow_state\x18\n \x01(\x0e\x32\x37.storage_onestore_v3_bytes.CompositeIndex.WorkflowStateB\x02\x18\x01\x12\x19\n\rerror_message\x18\x0b \x01(\tB\x02\x18\x01\x12\'\n\x14only_use_if_required\x18\x06 \x01(\x08:\x05\x66\x61lseB\x02\x18\x01\x12!\n\x0e\x64isabled_index\x18\t \x01(\x08:\x05\x66\x61lseB\x02\x18\x01\x12\'\n\x1f\x64\x65precated_read_division_family\x18\x07 \x03(\t\x12(\n deprecated_write_division_family\x18\x08 \x01(\t\"?\n\x05State\x12\x0e\n\nWRITE_ONLY\x10\x01\x12\x0e\n\nREAD_WRITE\x10\x02\x12\x0b\n\x07\x44\x45LETED\x10\x03\x12\t\n\x05\x45RROR\x10\x04\"7\n\rWorkflowState\x12\x0b\n\x07PENDING\x10\x01\x12\n\n\x06\x41\x43TIVE\x10\x02\x12\r\n\tCOMPLETED\x10\x03\"w\n\x10SearchIndexEntry\x12\x10\n\x08index_id\x18\x01 \x02(\x03\x12\x1d\n\x15write_division_family\x18\x02 \x02(\t\x12\x18\n\x10\x66ingerprint_1999\x18\x03 \x01(\x06\x12\x18\n\x10\x66ingerprint_2011\x18\x04 \x01(\x06\"\x98\x02\n\x0cIndexPostfix\x12G\n\x0bindex_value\x18\x01 \x03(\x0b\x32\x32.storage_onestore_v3_bytes.IndexPostfix.IndexValue\x12\x31\n\x03key\x18\x02 \x01(\x0b\x32$.storage_onestore_v3_bytes.Reference\x12\x14\n\x06\x62\x65\x66ore\x18\x03 \x01(\x08:\x04true\x12\x18\n\x10\x62\x65\x66ore_ascending\x18\x04 \x01(\x08\x1a\\\n\nIndexValue\x12\x15\n\rproperty_name\x18\x01 \x02(\t\x12\x37\n\x05value\x18\x02 \x02(\x0b\x32(.storage_onestore_v3_bytes.PropertyValue\"L\n\rIndexPosition\x12\x0b\n\x03key\x18\x01 \x01(\x0c\x12\x14\n\x06\x62\x65\x66ore\x18\x02 \x01(\x08:\x04true\x12\x18\n\x10\x62\x65\x66ore_ascending\x18\x03 \x01(\x08\x42\x45\n\x1e\x63om.google.storage.onestore.v3B\x0eOnestoreEntityZ\x13storage_onestore_v3')



_PROPERTYVALUE = DESCRIPTOR.message_types_by_name['PropertyValue']
_PROPERTYVALUE_POINTVALUE = _PROPERTYVALUE.nested_types_by_name['PointValue']
_PROPERTYVALUE_USERVALUE = _PROPERTYVALUE.nested_types_by_name['UserValue']
_PROPERTYVALUE_REFERENCEVALUE = _PROPERTYVALUE.nested_types_by_name['ReferenceValue']
_PROPERTYVALUE_REFERENCEVALUE_PATHELEMENT = _PROPERTYVALUE_REFERENCEVALUE.nested_types_by_name['PathElement']
_PROPERTY = DESCRIPTOR.message_types_by_name['Property']
_PATH = DESCRIPTOR.message_types_by_name['Path']
_PATH_ELEMENT = _PATH.nested_types_by_name['Element']
_REFERENCE = DESCRIPTOR.message_types_by_name['Reference']
_USER = DESCRIPTOR.message_types_by_name['User']
_ENTITYPROTO = DESCRIPTOR.message_types_by_name['EntityProto']
_ENTITYMETADATA = DESCRIPTOR.message_types_by_name['EntityMetadata']
_ENTITYSUMMARY = DESCRIPTOR.message_types_by_name['EntitySummary']
_ENTITYSUMMARY_PROPERTYSUMMARY = _ENTITYSUMMARY.nested_types_by_name['PropertySummary']
_COMPOSITEPROPERTY = DESCRIPTOR.message_types_by_name['CompositeProperty']
_INDEX = DESCRIPTOR.message_types_by_name['Index']
_INDEX_PROPERTY = _INDEX.nested_types_by_name['Property']
_COMPOSITEINDEX = DESCRIPTOR.message_types_by_name['CompositeIndex']
_SEARCHINDEXENTRY = DESCRIPTOR.message_types_by_name['SearchIndexEntry']
_INDEXPOSTFIX = DESCRIPTOR.message_types_by_name['IndexPostfix']
_INDEXPOSTFIX_INDEXVALUE = _INDEXPOSTFIX.nested_types_by_name['IndexValue']
_INDEXPOSITION = DESCRIPTOR.message_types_by_name['IndexPosition']
_PROPERTY_MEANING = _PROPERTY.enum_types_by_name['Meaning']
_ENTITYPROTO_KIND = _ENTITYPROTO.enum_types_by_name['Kind']
_INDEX_PROPERTY_DIRECTION = _INDEX_PROPERTY.enum_types_by_name['Direction']
_INDEX_PROPERTY_MODE = _INDEX_PROPERTY.enum_types_by_name['Mode']
_INDEX_VERSION = _INDEX.enum_types_by_name['Version']
_COMPOSITEINDEX_STATE = _COMPOSITEINDEX.enum_types_by_name['State']
_COMPOSITEINDEX_WORKFLOWSTATE = _COMPOSITEINDEX.enum_types_by_name['WorkflowState']
PropertyValue = _reflection.GeneratedProtocolMessageType('PropertyValue', (_message.Message,), {

  'PointValue' : _reflection.GeneratedProtocolMessageType('PointValue', (_message.Message,), {
    'DESCRIPTOR' : _PROPERTYVALUE_POINTVALUE,
    '__module__' : 'google.appengine.datastore.entity_bytes_pb2'

    })
  ,

  'UserValue' : _reflection.GeneratedProtocolMessageType('UserValue', (_message.Message,), {
    'DESCRIPTOR' : _PROPERTYVALUE_USERVALUE,
    '__module__' : 'google.appengine.datastore.entity_bytes_pb2'

    })
  ,

  'ReferenceValue' : _reflection.GeneratedProtocolMessageType('ReferenceValue', (_message.Message,), {

    'PathElement' : _reflection.GeneratedProtocolMessageType('PathElement', (_message.Message,), {
      'DESCRIPTOR' : _PROPERTYVALUE_REFERENCEVALUE_PATHELEMENT,
      '__module__' : 'google.appengine.datastore.entity_bytes_pb2'

      })
    ,
    'DESCRIPTOR' : _PROPERTYVALUE_REFERENCEVALUE,
    '__module__' : 'google.appengine.datastore.entity_bytes_pb2'

    })
  ,
  'DESCRIPTOR' : _PROPERTYVALUE,
  '__module__' : 'google.appengine.datastore.entity_bytes_pb2'

  })
_sym_db.RegisterMessage(PropertyValue)
_sym_db.RegisterMessage(PropertyValue.PointValue)
_sym_db.RegisterMessage(PropertyValue.UserValue)
_sym_db.RegisterMessage(PropertyValue.ReferenceValue)
_sym_db.RegisterMessage(PropertyValue.ReferenceValue.PathElement)

Property = _reflection.GeneratedProtocolMessageType('Property', (_message.Message,), {
  'DESCRIPTOR' : _PROPERTY,
  '__module__' : 'google.appengine.datastore.entity_bytes_pb2'

  })
_sym_db.RegisterMessage(Property)

Path = _reflection.GeneratedProtocolMessageType('Path', (_message.Message,), {

  'Element' : _reflection.GeneratedProtocolMessageType('Element', (_message.Message,), {
    'DESCRIPTOR' : _PATH_ELEMENT,
    '__module__' : 'google.appengine.datastore.entity_bytes_pb2'

    })
  ,
  'DESCRIPTOR' : _PATH,
  '__module__' : 'google.appengine.datastore.entity_bytes_pb2'

  })
_sym_db.RegisterMessage(Path)
_sym_db.RegisterMessage(Path.Element)

Reference = _reflection.GeneratedProtocolMessageType('Reference', (_message.Message,), {
  'DESCRIPTOR' : _REFERENCE,
  '__module__' : 'google.appengine.datastore.entity_bytes_pb2'

  })
_sym_db.RegisterMessage(Reference)

User = _reflection.GeneratedProtocolMessageType('User', (_message.Message,), {
  'DESCRIPTOR' : _USER,
  '__module__' : 'google.appengine.datastore.entity_bytes_pb2'

  })
_sym_db.RegisterMessage(User)

EntityProto = _reflection.GeneratedProtocolMessageType('EntityProto', (_message.Message,), {
  'DESCRIPTOR' : _ENTITYPROTO,
  '__module__' : 'google.appengine.datastore.entity_bytes_pb2'

  })
_sym_db.RegisterMessage(EntityProto)

EntityMetadata = _reflection.GeneratedProtocolMessageType('EntityMetadata', (_message.Message,), {
  'DESCRIPTOR' : _ENTITYMETADATA,
  '__module__' : 'google.appengine.datastore.entity_bytes_pb2'

  })
_sym_db.RegisterMessage(EntityMetadata)

EntitySummary = _reflection.GeneratedProtocolMessageType('EntitySummary', (_message.Message,), {

  'PropertySummary' : _reflection.GeneratedProtocolMessageType('PropertySummary', (_message.Message,), {
    'DESCRIPTOR' : _ENTITYSUMMARY_PROPERTYSUMMARY,
    '__module__' : 'google.appengine.datastore.entity_bytes_pb2'

    })
  ,
  'DESCRIPTOR' : _ENTITYSUMMARY,
  '__module__' : 'google.appengine.datastore.entity_bytes_pb2'

  })
_sym_db.RegisterMessage(EntitySummary)
_sym_db.RegisterMessage(EntitySummary.PropertySummary)

CompositeProperty = _reflection.GeneratedProtocolMessageType('CompositeProperty', (_message.Message,), {
  'DESCRIPTOR' : _COMPOSITEPROPERTY,
  '__module__' : 'google.appengine.datastore.entity_bytes_pb2'

  })
_sym_db.RegisterMessage(CompositeProperty)

Index = _reflection.GeneratedProtocolMessageType('Index', (_message.Message,), {

  'Property' : _reflection.GeneratedProtocolMessageType('Property', (_message.Message,), {
    'DESCRIPTOR' : _INDEX_PROPERTY,
    '__module__' : 'google.appengine.datastore.entity_bytes_pb2'

    })
  ,
  'DESCRIPTOR' : _INDEX,
  '__module__' : 'google.appengine.datastore.entity_bytes_pb2'

  })
_sym_db.RegisterMessage(Index)
_sym_db.RegisterMessage(Index.Property)

CompositeIndex = _reflection.GeneratedProtocolMessageType('CompositeIndex', (_message.Message,), {
  'DESCRIPTOR' : _COMPOSITEINDEX,
  '__module__' : 'google.appengine.datastore.entity_bytes_pb2'

  })
_sym_db.RegisterMessage(CompositeIndex)

SearchIndexEntry = _reflection.GeneratedProtocolMessageType('SearchIndexEntry', (_message.Message,), {
  'DESCRIPTOR' : _SEARCHINDEXENTRY,
  '__module__' : 'google.appengine.datastore.entity_bytes_pb2'

  })
_sym_db.RegisterMessage(SearchIndexEntry)

IndexPostfix = _reflection.GeneratedProtocolMessageType('IndexPostfix', (_message.Message,), {

  'IndexValue' : _reflection.GeneratedProtocolMessageType('IndexValue', (_message.Message,), {
    'DESCRIPTOR' : _INDEXPOSTFIX_INDEXVALUE,
    '__module__' : 'google.appengine.datastore.entity_bytes_pb2'

    })
  ,
  'DESCRIPTOR' : _INDEXPOSTFIX,
  '__module__' : 'google.appengine.datastore.entity_bytes_pb2'

  })
_sym_db.RegisterMessage(IndexPostfix)
_sym_db.RegisterMessage(IndexPostfix.IndexValue)

IndexPosition = _reflection.GeneratedProtocolMessageType('IndexPosition', (_message.Message,), {
  'DESCRIPTOR' : _INDEXPOSITION,
  '__module__' : 'google.appengine.datastore.entity_bytes_pb2'

  })
_sym_db.RegisterMessage(IndexPosition)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\036com.google.storage.onestore.v3B\016OnestoreEntityZ\023storage_onestore_v3'
  _COMPOSITEINDEX.fields_by_name['workflow_state']._options = None
  _COMPOSITEINDEX.fields_by_name['workflow_state']._serialized_options = b'\030\001'
  _COMPOSITEINDEX.fields_by_name['error_message']._options = None
  _COMPOSITEINDEX.fields_by_name['error_message']._serialized_options = b'\030\001'
  _COMPOSITEINDEX.fields_by_name['only_use_if_required']._options = None
  _COMPOSITEINDEX.fields_by_name['only_use_if_required']._serialized_options = b'\030\001'
  _COMPOSITEINDEX.fields_by_name['disabled_index']._options = None
  _COMPOSITEINDEX.fields_by_name['disabled_index']._serialized_options = b'\030\001'
  _PROPERTYVALUE._serialized_start=77
  _PROPERTYVALUE._serialized_end=822
  _PROPERTYVALUE_POINTVALUE._serialized_start=403
  _PROPERTYVALUE_POINTVALUE._serialized_end=437
  _PROPERTYVALUE_USERVALUE._serialized_start=440
  _PROPERTYVALUE_USERVALUE._serialized_end=604
  _PROPERTYVALUE_REFERENCEVALUE._serialized_start=607
  _PROPERTYVALUE_REFERENCEVALUE._serialized_end=822
  _PROPERTYVALUE_REFERENCEVALUE_PATHELEMENT._serialized_start=769
  _PROPERTYVALUE_REFERENCEVALUE_PATHELEMENT._serialized_end=822
  _PROPERTY._serialized_start=825
  _PROPERTY._serialized_end=1409
  _PROPERTY_MEANING._serialized_start=1068
  _PROPERTY_MEANING._serialized_end=1409
  _PATH._serialized_start=1411
  _PATH._serialized_end=1526
  _PATH_ELEMENT._serialized_start=1477
  _PATH_ELEMENT._serialized_end=1526
  _REFERENCE._serialized_start=1528
  _REFERENCE._serialized_end=1640
  _USER._serialized_start=1643
  _USER._serialized_end=1802
  _ENTITYPROTO._serialized_start=1805
  _ENTITYPROTO._serialized_end=2217
  _ENTITYPROTO_KIND._serialized_start=2165
  _ENTITYPROTO_KIND._serialized_end=2217
  _ENTITYMETADATA._serialized_start=2219
  _ENTITYMETADATA._serialized_end=2285
  _ENTITYSUMMARY._serialized_start=2288
  _ENTITYSUMMARY._serialized_end=2475
  _ENTITYSUMMARY_PROPERTYSUMMARY._serialized_start=2391
  _ENTITYSUMMARY_PROPERTYSUMMARY._serialized_end=2475
  _COMPOSITEPROPERTY._serialized_start=2477
  _COMPOSITEPROPERTY._serialized_end=2529
  _INDEX._serialized_start=2532
  _INDEX._serialized_end=3134
  _INDEX_PROPERTY._serialized_start=2738
  _INDEX_PROPERTY._serialized_end=3074
  _INDEX_PROPERTY_DIRECTION._serialized_start=2939
  _INDEX_PROPERTY_DIRECTION._serialized_end=3008
  _INDEX_PROPERTY_MODE._serialized_start=3010
  _INDEX_PROPERTY_MODE._serialized_end=3074
  _INDEX_VERSION._serialized_start=3076
  _INDEX_VERSION._serialized_end=3134
  _COMPOSITEINDEX._serialized_start=3137
  _COMPOSITEINDEX._serialized_end=3713
  _COMPOSITEINDEX_STATE._serialized_start=3593
  _COMPOSITEINDEX_STATE._serialized_end=3656
  _COMPOSITEINDEX_WORKFLOWSTATE._serialized_start=3658
  _COMPOSITEINDEX_WORKFLOWSTATE._serialized_end=3713
  _SEARCHINDEXENTRY._serialized_start=3715
  _SEARCHINDEXENTRY._serialized_end=3834
  _INDEXPOSTFIX._serialized_start=3837
  _INDEXPOSTFIX._serialized_end=4117
  _INDEXPOSTFIX_INDEXVALUE._serialized_start=4025
  _INDEXPOSTFIX_INDEXVALUE._serialized_end=4117
  _INDEXPOSITION._serialized_start=4119
  _INDEXPOSITION._serialized_end=4195




#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""A Cloud Datastore stub that connects to a remote Datastore service."""

from google.appengine.api import apiproxy_rpc
from google.appengine.api import apiproxy_stub
from google.appengine.api import datastore_errors
from google.appengine.datastore import datastore_pbs
from google.appengine.datastore import datastore_rpc


class CloudDatastoreV1RemoteStub(apiproxy_stub.APIProxyStub):
  """A stub for calling Cloud Datastore via the Cloud Datastore API."""

  def __init__(self, datastore):
    """Constructs a new Cloud Datastore stub.

    Args:
      datastore: A googledatastore.Datastore object.
    """
    self._datastore = datastore

  def MakeSyncCall(self, service, call, request, response):
    assert service == 'cloud_datastore_v1'


    call = (call[0:1].lower() + call[1:])

    call = call.encode('utf-8')


    try:
      response.CopyFrom(
          self._datastore._call_method(call, request, response.__class__))
    except datastore_pbs.googledatastore.RPCError as e:
      raise datastore_rpc._DatastoreExceptionFromCanonicalErrorCodeAndDetail(
          e.code, e.message)
    except Exception as e:
      raise datastore_errors.InternalError(e)

  def CreateRPC(self):
    return apiproxy_rpc.RPC(stub=self)



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""App identity stub service implementation.

This service behaves the same as the production service, except using
constant values instead of app-specific values:
* Signing key is constant; in production this rotates.
* Public key is constant; in production this varies per app.
* Service account name is constant; in production this varies per app.
"""






import logging

from google.auth import exceptions

from google.appengine.api.app_identity import app_identity_defaultcredentialsbased_stub as ai_default_stub
from google.appengine.api.app_identity import app_identity_keybased_stub
from google.appengine.api.app_identity import app_identity_stub_base


APP_SERVICE_ACCOUNT_NAME = app_identity_stub_base.APP_SERVICE_ACCOUNT_NAME


class AppIdentityServiceStub(app_identity_stub_base.AppIdentityServiceStubBase):
  """A stub for the AppIdentityService API for offline development.

  Provides stub functions which allow a developer to test integration before
  deployment.

  Automatically creates appropriate stub (default credentials or key based)
  using `Create()`.
  """
  THREADSAFE = True

  @staticmethod
  def Create(email_address=None, private_key_path=None, oauth_url=None):
    if email_address:
      logging.debug('Using the KeyBasedAppIdentityServiceStub.')
      return app_identity_keybased_stub.KeyBasedAppIdentityServiceStub(
          email_address=email_address,
          private_key_path=private_key_path,
          oauth_url=oauth_url)
    else:
      try:
        dc = ai_default_stub.DefaultCredentialsBasedAppIdentityServiceStub()
        logging.debug('Successfully loaded Application Default Credentials.')
        return dc
      except exceptions.DefaultCredentialsError as error:
        if not str(error).startswith(
            'Could not automatically determine credentials.'):

          logging.warning('An exception has been encountered when attempting '
                          'to use Application Default Credentials: %s'
                          '. Falling back on dummy AppIdentityServiceStub.',
                          str(error))
        return AppIdentityServiceStub()



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Utilities for generating and updating index.yaml."""

import logging
import os

import six
from six.moves import zip

from google.appengine.api import apiproxy_stub_map
from google.appengine.api import validation
from google.appengine.api import yaml_errors
from google.appengine.datastore import datastore_index
from google.appengine.datastore import datastore_index_xml








__all__ = ['GenerateIndexFromHistory',
           'IndexYamlUpdater',
          ]

logger = logging.getLogger('google.appengine.api.stubs.datastore')


AUTO_MARKER = '\n# AUTOGENERATED\n'


AUTO_COMMENT = '''
# This index.yaml is automatically updated whenever the dev_appserver
# detects that a new type of query is run.  If you want to manage the
# index.yaml file manually, remove the above marker line (the line
# saying "# AUTOGENERATED").  If you want to manage some indexes
# manually, move them above the marker line.  The index.yaml file is
# automatically uploaded to the admin console when you next deploy
# your application using appcfg.py.
'''


def GenerateIndexDictFromHistory(query_history,
                                 all_indexes=None, manual_indexes=None):
  """Generate a dict of automatic index entries from the query history.

  Args:
    query_history: Query history, a dict mapping datastore_pb.Query to a count
      of the number of times that query has been issued.
    all_indexes: Optional datastore_index.IndexDefinitions instance
      representing all the indexes found in the input file.  May be None.
    manual_indexes: Optional datastore_index.IndexDefinitions instance
      containing indexes for which we should not generate output.  May be None.

  Returns:
    A dict where each key is a tuple (kind, ancestor, properties) and the
      corresponding value is a count of the number of times that query has been
      issued. The dict contains no entries for keys that appear in manual_keys.
      In the tuple, "properties" is itself a tuple of tuples, where each
      contained tuple is (name, direction), with "name" being a string and
      "direction" being datastore_index.ASCENDING or .DESCENDING.
  """



  all_keys = datastore_index.IndexDefinitionsToKeys(all_indexes)
  manual_keys = datastore_index.IndexDefinitionsToKeys(manual_indexes)


  indexes = dict((key, 0) for key in all_keys - manual_keys)


  for query, count in six.iteritems(query_history):
    required, kind, ancestor, props = (
        datastore_index.CompositeIndexForQuery(query))
    if required:
      props = datastore_index.GetRecommendedIndexProperties(props)
      spec = (kind, ancestor, props)




      if not any(_IndexSpecSatisfies(spec, k) for k in manual_keys):
        _UpdateGeneratedIndexes(spec, count, indexes)

  return indexes


def _UpdateGeneratedIndexes(spec, count, indexes):
  """Updates the set of generated indexes to cover given query requirements.

  It may add "spec" to the "indexes" dict if the latter does not
  already have an index suitable for serving the query represented by
  "spec".  (It may even replace an entry already in the dict if that
  gives optimal coverage.)

  Args:
    spec: specification of index requirements (in "key" form) for a
      query executed in the stub.
    count: number of times the query was executed
    indexes: dict containing other already-generated index "keys" and
      their counts.

  No return value; instead it Updates "indexes" in place as necessary.

  """




  for index in indexes:
    if _IndexSpecSatisfies(spec, index):
      indexes[index] += count
      return



  for index in indexes:
    if _IndexSpecSatisfies(index, spec):
      indexes[spec] = indexes[index] + count
      del indexes[index]
      return


  indexes[spec] = count


def _IndexSpecSatisfies(spec, candidate):
  """Determines whether candidate index can serve the query given by spec."""
  (spec_kind, spec_ancestor, spec_props) = spec
  (candidate_kind, candidate_ancestor, candidate_props) = candidate
  if (spec_kind, spec_ancestor) != (candidate_kind, candidate_ancestor):
    return False
  if len(spec_props) != len(candidate_props):
    return False
  return all(_PropSatisfies(s, c)
             for (s, c) in zip(spec_props, candidate_props))


def _PropSatisfies(spec, candidate):
  """Determines whether candidate property meets requirements given by spec."""


















  if spec.name != candidate.name or spec.mode != candidate.mode:
    return False

  return True if spec.direction is None else (
      spec.direction == candidate.direction)


def GenerateIndexFromHistory(query_history,
                             all_indexes=None, manual_indexes=None):
  """Generate most of the text for index.yaml from the query history.

  Args:
    query_history: Query history, a dict mapping datastore_pb.Query to a count
      of the number of times that query has been issued.
    all_indexes: Optional datastore_index.IndexDefinitions instance
      representing all the indexes found in the input file.  May be None.
    manual_indexes: Optional datastore_index.IndexDefinitions instance
      containing indexes for which we should not generate output.  May be None.

  Returns:
    A string representation that can safely be appended to an existing
    index.yaml file. Returns the empty string if it would generate no output.
  """
  indexes = GenerateIndexDictFromHistory(
      query_history, all_indexes, manual_indexes)

  if not indexes:
    return ''




  res = []
  for (kind, ancestor, props), _ in sorted(six.iteritems(indexes)):

    res.append('')
    res.append(datastore_index.IndexYamlForQuery(kind, ancestor, props))

  res.append('')
  return '\n'.join(res)


class EmptyIndexFileError(Exception):
  """Raised when index.yaml is empty."""


class IndexYamlUpdater(object):
  """Helper class for updating index.yaml.

  This class maintains some state about the query history and the
  index.yaml file in order to minimize the number of times index.yaml
  is actually overwritten.
  """


  index_yaml_is_manual = False
  index_yaml_mtime = None
  last_history_size = 0

  def __init__(self, root_path):
    """Constructor.

    Args:
      root_path: Path to the app's root directory.
    """
    self.root_path = root_path

  def UpdateIndexConfig(self):
    self.UpdateIndexYaml()

  def UpdateIndexYaml(self, openfile=open):
    """Update index.yaml.

    Args:
      openfile: Used for dependency injection.

    We only ever write to index.yaml if either:
    - it doesn't exist yet; or
    - it contains an 'AUTOGENERATED' comment.

    All indexes *before* the AUTOGENERATED comment will be written
    back unchanged.  All indexes *after* the AUTOGENERATED comment
    will be updated with the latest query counts (query counts are
    reset by --clear_datastore).  Indexes that aren't yet in the file
    will be appended to the AUTOGENERATED section.

    We keep track of some data in order to avoid doing repetitive work:
    - if index.yaml is fully manual, we keep track of its mtime to
      avoid parsing it over and over;
    - we keep track of the number of keys in the history dict since
      the last time we updated index.yaml (or decided there was
      nothing to update).
    """





    index_yaml_file = os.path.join(self.root_path, 'index.yaml')


    try:
      index_yaml_mtime = os.path.getmtime(index_yaml_file)
    except os.error:
      index_yaml_mtime = None


    index_yaml_changed = (index_yaml_mtime != self.index_yaml_mtime)
    self.index_yaml_mtime = index_yaml_mtime


    datastore_stub = apiproxy_stub_map.apiproxy.GetStub('datastore_v3')
    query_ci_history_len = datastore_stub._QueryCompositeIndexHistoryLength()
    history_changed = (query_ci_history_len != self.last_history_size)
    self.last_history_size = query_ci_history_len


    if not (index_yaml_changed or history_changed):
      logger.debug('No need to update index.yaml')
      return


    if self.index_yaml_is_manual and not index_yaml_changed:
      logger.debug('Will not update manual index.yaml')
      return


    if index_yaml_mtime is None:
      index_yaml_data = None
    else:
      try:



        fh = openfile(index_yaml_file, 'rU')
      except IOError:
        index_yaml_data = None
      else:
        try:
          index_yaml_data = fh.read()
          if not index_yaml_data:
            raise EmptyIndexFileError(
                'The index yaml file is empty. The file should at least have '
                'an empty "indexes:" block.')
        finally:
          fh.close()


    self.index_yaml_is_manual = (index_yaml_data is not None and
                                 AUTO_MARKER not in index_yaml_data)
    if self.index_yaml_is_manual:
      logger.info('Detected manual index.yaml, will not update')
      return



    if index_yaml_data is None:
      all_indexes = None
    else:
      try:
        all_indexes = datastore_index.ParseIndexDefinitions(index_yaml_data)
      except yaml_errors.EventListenerError as e:

        logger.error('Error parsing %s:\n%s', index_yaml_file, e)
        return
      except Exception as err:

        logger.error('Error parsing %s:\n%s.%s: %s', index_yaml_file,
                     err.__class__.__module__, err.__class__.__name__, err)
        return


    if index_yaml_data is None:
      manual_part, prev_automatic_part = 'indexes:\n', ''
      manual_indexes = None
    else:
      manual_part, prev_automatic_part = index_yaml_data.split(AUTO_MARKER, 1)
      if prev_automatic_part.startswith(AUTO_COMMENT):
        prev_automatic_part = prev_automatic_part[len(AUTO_COMMENT):]

      try:
        manual_indexes = datastore_index.ParseIndexDefinitions(manual_part)
      except Exception as err:
        logger.error('Error parsing manual part of %s: %s',
                     index_yaml_file, err)
        return


    automatic_part = GenerateIndexFromHistory(datastore_stub.QueryHistory(),
                                              all_indexes, manual_indexes)



    if (index_yaml_mtime is None and automatic_part == '' or
        automatic_part == prev_automatic_part):
      logger.debug('No need to update index.yaml')
      return


    try:
      fh = openfile(index_yaml_file, 'w')
    except IOError as err:
      logger.error('Can\'t write index.yaml: %s', err)
      return


    try:
      logger.info('Updating %s', index_yaml_file)
      fh.write(manual_part)
      fh.write(AUTO_MARKER)
      fh.write(AUTO_COMMENT)
      fh.write(automatic_part)
    finally:
      fh.close()


    try:
      self.index_yaml_mtime = os.path.getmtime(index_yaml_file)
    except os.error as err:
      logger.error('Can\'t stat index.yaml we just wrote: %s', err)
      self.index_yaml_mtime = None


class DatastoreIndexesAutoXmlUpdater(object):
  """Helper class for updating datastore-indexes-auto.xml.

  This class maintains some state about the query history and the
  datastore-indexes.xml and datastore-indexes-auto.xml files in order to
  minimize the number of times datastore-indexes-auto.xml is rewritten.
  """



  auto_generated = True
  datastore_indexes_xml = None
  datastore_indexes_xml_mtime = None
  datastore_indexes_auto_xml_mtime = None
  last_history_size = 0

  def __init__(self, root_path):
    self.root_path = root_path

  def UpdateIndexConfig(self):
    self.UpdateDatastoreIndexesAutoXml()

  def UpdateDatastoreIndexesAutoXml(self, openfile=open):
    """Update datastore-indexes-auto.xml if appropriate."""




    datastore_indexes_xml_file = os.path.join(
        self.root_path, 'WEB-INF', 'datastore-indexes.xml')
    try:
      datastore_indexes_xml_mtime = os.path.getmtime(datastore_indexes_xml_file)
    except os.error:
      datastore_indexes_xml_mtime = None
    if datastore_indexes_xml_mtime != self.datastore_indexes_xml_mtime:
      self.datastore_indexes_xml_mtime = datastore_indexes_xml_mtime
      if self.datastore_indexes_xml_mtime:
        with openfile(datastore_indexes_xml_file) as f:
          self.datastore_indexes_xml = f.read()
          self.auto_generated = datastore_index_xml.IsAutoGenerated(
              self.datastore_indexes_xml)
      else:
        self.auto_generated = True
        self.datastore_indexes_xml = None

    if not self.auto_generated:
      logger.debug('Detected <datastore-indexes autoGenerated="false">,'
                   ' will not update datastore-indexes-auto.xml')
      return


    datastore_stub = apiproxy_stub_map.apiproxy.GetStub('datastore_v3')
    query_ci_history_len = datastore_stub._QueryCompositeIndexHistoryLength()
    history_changed = (query_ci_history_len != self.last_history_size)
    self.last_history_size = query_ci_history_len
    if not history_changed:
      logger.debug('No need to update datastore-indexes-auto.xml')
      return

    datastore_indexes_auto_xml_file = os.path.join(
        self.root_path, 'WEB-INF', 'appengine-generated',
        'datastore-indexes-auto.xml')
    try:
      with open(datastore_indexes_auto_xml_file) as f:
        datastore_indexes_auto_xml = f.read()
    except IOError as err:
      datastore_indexes_auto_xml = None

    if self.datastore_indexes_xml:
      try:
        manual_index_definitions = (
            datastore_index_xml.IndexesXmlToIndexDefinitions(
                self.datastore_indexes_xml))
      except validation.ValidationError as e:
        logger.error('Error parsing %s: %s',
                     datastore_indexes_xml_file, e)
        return
    else:
      manual_index_definitions = datastore_index.IndexDefinitions(indexes=[])

    if datastore_indexes_auto_xml:
      try:
        prev_auto_index_definitions = (
            datastore_index_xml.IndexesXmlToIndexDefinitions(
                datastore_indexes_auto_xml))
      except validation.ValidationError as e:
        logger.error('Error parsing %s: %s',
                     datastore_indexes_auto_xml_file, e)
        return
    else:
      prev_auto_index_definitions = datastore_index.IndexDefinitions(indexes=[])

    all_index_definitions = datastore_index.IndexDefinitions(
        indexes=(manual_index_definitions.indexes +
                 prev_auto_index_definitions.indexes))
    query_history = datastore_stub.QueryHistory()
    auto_index_dict = GenerateIndexDictFromHistory(
        query_history, all_index_definitions, manual_index_definitions)
    auto_indexes, counts = self._IndexesFromIndexDict(auto_index_dict)
    auto_index_definitions = datastore_index.IndexDefinitions(
        indexes=auto_indexes)
    if auto_index_definitions == prev_auto_index_definitions:
      return

    try:
      appengine_generated = os.path.dirname(datastore_indexes_auto_xml_file)
      if not os.path.exists(appengine_generated):
        os.mkdir(appengine_generated)
      with open(datastore_indexes_auto_xml_file, 'w') as f:
        f.write(self._IndexXmlFromIndexes(auto_indexes, counts))
    except os.error as err:
      logger.error(
          'Could not update %s: %s', datastore_indexes_auto_xml_file, err)

  def _IndexesFromIndexDict(self, index_dict):
    """Convert a query dictionary into the corresponding required indexes.

    Args:
      index_dict: Index usage history, a dict mapping composite index
        descriptors to a count of the number of times that queries
        needing such an index have been executed

    Returns:
      a tuple (indexes, counts) where indexes and counts are lists of the same
      size, with each entry in indexes being a datastore_index.Index and each
      entry in indexes being the count of the number of times the corresponding
      query appeared in the history.
    """
    indexes = []
    counts = []
    for (kind, ancestor, props), count in sorted(six.iteritems(index_dict)):
      properties = []
      for prop in props:
        if prop.direction is None:
          direction = None
        else:
          direction = (
              'desc' if prop.direction == datastore_index.DESCENDING else 'asc')
        mode = (
            'geospatial' if prop.mode == datastore_index.GEOSPATIAL else None)
        properties.append(datastore_index.Property(
            name=prop.name, direction=direction, mode=mode))

      indexes.append(datastore_index.Index(
          kind=kind, ancestor=bool(ancestor), properties=properties))
      counts.append(count)

    return indexes, counts

  def _IndexXmlFromIndexes(self, indexes, counts):
    """Create <datastore-indexes> XML for the given indexes and query counts.

    Args:
      indexes: a list of datastore_index.Index objects that are the required
        indexes.
      counts: a list of integers that are the corresponding counts.

    Returns:
      the corresponding XML, with root node <datastore-indexes>.
    """
    lines = ['<datastore-indexes>']
    for index, count in zip(indexes, counts):
      lines.append('  <!-- Used %d time%s in query history -->'
                   % (count, 's' if count != 1 else ''))
      kind, ancestor, props = datastore_index.IndexToKey(index)
      xml_fragment = datastore_index.IndexXmlForQuery(kind, ancestor, props)
      lines.append(xml_fragment)

    lines.append('</datastore-indexes>')
    return '\n'.join(lines) + '\n'
